///
//  Callback v 1.6
//  Russell Lowke, August 23rd 2010
//  for use with ActionScript 3 (AS3, Flex SDK 3.2, Flash Player 9)
// 
//  Copyright (c) 2009-2010 Russell Lowke
//  see http://www.lowkemedia.com for more information
//
//  Permission is hereby granted, free of charge, to any person obtaining a 
//  copy of this software and associated documentation files (the "Software"), 
//  to deal in the Software without restriction, including without limitation 
//  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
//  and/or sell copies of the Software, and to permit persons to whom the 
//  Software is furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in 
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
//  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
//  IN THE SOFTWARE. 
//
//

//
//  Callback class
//
//  Callback listens for an event from any EventDispatcher object and, just 
//  once, triggers a function with arguments when that event is dispatched. 
//  It's an easy way to piggyback additional arguments onto an event listener, 
//  essentially turning an event into a function callback with arguments.
//
//  Note: Callback is attached to the object being listened to, if that 
//  object gets cleared from memory then so too is the Callback instance.
//
//
//  Usage:
//
//      import com.util.Callback;
//      Callback.call(spriteObj, Event.ENTER_FRAME, funct, ... args);
//
//  Where the function funct will be called and any args passed to it   
//  when, in this case, the EventDispatcher spriteObj next dispatches an 
//  ENTER_FRAME event.
//
//
//  There is also a callWithEvt() method,
//
//      Callback.callWithEvt(spriteObj, Event.ENTER_FRAME, funct, ... args);
//
//  Where the callback function is passed the dispatched evt as the 1st
//  argument, then any additional args.
//
//  
//  If you want the callback to keep listening for the event and trigger 
//  every time the event is dispatched use,
//
//      Callback.listenFor(spriteObj, Event.ENTER_FRAME, funct, ... args);
//
//  Where the function funct will be called and any args passed to it  
//  every time the EventDispatcher spriteObj dispatches an ENTER_FRAME event.
//


package com.lowke.utils.as3 
{
    import flash.events.Event;
    import flash.events.EventDispatcher;
    
    public class Callback 
    {   
        //
        // member variables
        private var _obj:EventDispatcher;                   // dispatcher being listened to
        private var _evtString:String;                      // event string being listened for
        private var _passEvt:Boolean;                       // if true then the evt is passed along with any args
        private var _persist:Boolean;                       // if true then the Callback is not automatically removed
        private var _funct:Function;                        // function to be called
        private var _args:Array;                            // arguments to be sent to function
        
        
        //
        // constructor
        public function Callback(obj:EventDispatcher,
                                 evtString:String,
                                 passEvt:Boolean = false,
                                 persist:Boolean = false,
                                 funct:Function = null,
                                 args:Array = null):void 
        {
            _obj        = obj;
            _evtString  = evtString;
            _passEvt    = passEvt;
            _persist    = persist;
            _funct      = funct;
            if (! args) {
                _args = new Array();
            } else {
                _args = args;
            }
            
            if (_funct == null) {
                throw new Error("funct parameter is null. Callback must have a function to callback.");
            }
            
            if (_obj == null) {
                throw new Error("obj parameter is null. Callback must listen to an EventDispatcher object.");
            }
            
            // Note: this listener must be a hard listener and not soft
            //  as it is intended to be the only reference to the Callback instance
            _obj.addEventListener(_evtString, trigger);
        }
        
        public function trigger(evt:Event = null):void 
        {   
            if (! _persist) {
                close();
            }
            
            if (_passEvt) {
                _args = [evt].concat(_args);
            }
            _funct.apply(null, _args);
        }
        
        public function close():void 
        {   
            // remove the event listener, clearing the Callback.
            _obj.removeEventListener(_evtString, trigger);
        }
        
        
        
        //
        // accessors and mutators
        //
        
        public function get obj():EventDispatcher           { return _obj; }
        public function get evtString():String              { return _evtString; }
        public function get funct():Function                { return _funct; }
        
        
        
        //
        //  static helper methods
        //
        
        public static function call(obj:EventDispatcher, 
                                    evtString:String, 
                                    funct:Function, 
                                    ... args):Callback
        {
            return new Callback(obj, evtString, false, false, funct, args);
        }
        
        public static function callWithEvt(obj:EventDispatcher, 
                                           evtString:String, 
                                           funct:Function, 
                                           ... args):Callback
        {
            return new Callback(obj, evtString, true, false, funct, args);
        }
        
        public static function listenFor(obj:EventDispatcher, 
                                         evtString:String, 
                                         funct:Function, 
                                         ... args):Callback
        {
            return new Callback(obj, evtString, false, true, funct, args);
        }
        
        
        /** 
         * argsCall() is intended for internal callbacks where arguments are 
         * already packed into args Array.
         */
        public static function argsCall(obj:EventDispatcher, 
                                        evtString:String, 
                                        funct:Function, 
                                        args:Array):Callback
        {
            var callback:Callback;
            if (funct != null) {
                callback = new Callback(obj, evtString, false, false, funct, args);
            }
            return callback;
        }
        
        /** 
         * argsListenFor() is intended for internal callbacks where arguments are 
         * already packed into args Array.
         */
        public static function argsListenFor(obj:EventDispatcher, 
                                             evtString:String, 
                                             funct:Function, 
                                             args:Array):Callback
        {
            var callback:Callback;
            if (funct != null) {
                callback = new Callback(obj, evtString, false, true, funct, args);
            }
            return callback;
        }
    }
}