///
//  DoLater v 1.5.3
//  Russell Lowke, September 9th 2010
//  for use with ActionScript 3 (AS3, Flex SDK 3.2, Flash Player 9)
//
//  Copyright (c) 2008-2009 Russell Lowke
//  see http://www.lowkemedia.com for more information
//
//  Permission is hereby granted, free of charge, to any person obtaining a 
//  copy of this software and associated documentation files (the "Software"), 
//  to deal in the Software without restriction, including without limitation 
//  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
//  and/or sell copies of the Software, and to permit persons to whom the 
//  Software is furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in 
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
//  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
//  IN THE SOFTWARE. 
//
//

//
//  DoLater class
//
//  DoLater is used to trigger a callback after the frame has updated.
//  If you're looking for a callback after a specific time see Delay.as class.
//
//  Unlike traditional programming languages, ActionScript is executed within 
//  the Flash Player. Flash is based on the movie concept of a timeline, where 
//  symbols and code can be placed on various frames and Flash movies run at 
//  specified frame rates. Flash updates the visuals (color, object placement, 
//  size, etc.) at the end of the frame. This gives the ActionScript in the 
//  frame a chance to execute and set up the visual changes.
//  
//  The doLater() method, as defined in the Flex mx.core.UIObject, allows you to 
//  queue a function to be called after the frame has been updated. A function 
//  might need to be called on the next update to allow the rest of the code 
//  scheduled for the current update to be executed.
//
//  Paradoxically, regular Flash ActionScript 3 has no inbuilt doLater() function 
//  like the Flex UIObject component (Flex mx.core.UIObject). 
// 
//  This DoLater class fills that gap.
//
//
//  Usage:
//
//      import com.lowke.utils.as3.DoLater;
//      DoLater.doIt(funct, ... args);
//
//  Where the function funct will be called and args passed to it after the 
//  next Event.ENTER_FRAME, once the frame has updated and any actionscript on
//  the frame has been executed.
//
//
//  There is also a onEnterFrame() function,
//
//      DoLater.onEnterFrame(funct, ... args);
//
//  Where the function funct will be called and args passed to it on the next 
//  Event.ENTER_FRAME.
//
        
package com.lowke.utils.as3 
{
    import flash.display.Sprite;
    import flash.events.Event;

    public class DoLater 
    {    
        // a static _sprite is used to listen for ENTER_FRAME events,
        //  this is an assured way of getting access to stage.
        private static var _sprite:Sprite = new Sprite();
        
        //
        // member variables
        private var _funct:Function;        // function to be called later
        private var _args:Array;            // arguments to be sent to function
        private var _ready:Boolean;         // wait one ENTER_FRAME to ensure the whole frame has loaded
        
        //
        // constructor
        public function DoLater(funct:Function,
                                args:Array = null):void 
        {
            _ready = false;
            _funct = funct;
            _args  = args;
            
            // Note: this listener must be a hard listener and not soft
            //  as it is intended to be the only reference to the DoLater instance
            _sprite.addEventListener(Event.ENTER_FRAME, enterFrameEvt);
        }
        
        private function enterFrameEvt(evt:Event):void 
        {    
            if (_ready) {
                trigger();
                
            } else {
                
                // wait one ENTER_FRAME to ensure the frame has loaded
                //  and triggered all ActionScript on the frame.
                _ready = true;
            }
        }
        
        public function trigger():void 
        {   
            // remove the listener
            close();
            
            // call the function, passing into it args
            if (_funct != null) {
                _funct.apply(null, _args);
            }
        }
        
        public function close():void 
        {    
            // remove the event listener, clearing the DoLater.
            _sprite.removeEventListener(Event.ENTER_FRAME, enterFrameEvt);
        }
        
        public function set ready(val:Boolean):void 
        {
            _ready = val;
        }
        
        // usually you want the function to trigger after the
        //  frame has updated and all items/code on various MovieClips
        //  have been fully executed, so use doIt().
        public static function doIt(funct:Function, ... args):DoLater 
        {
            return new DoLater(funct, args);
        }
        
        // ...but sometimes it is advantageous to have the function
        //  trigger before on the next frame, so use onEnterFrame().
        public static function onEnterFrame(funct:Function, ... args):DoLater 
        {
            var doLaterObj:DoLater = new DoLater(funct, args);
            doLaterObj.ready = true;
            return doLaterObj;
        }
    }
}