///
//  AssetLoader v 2.6.2
//  Russell Lowke, February 9th 2011
//
//  Copyright (c) 2008-2011 Russell Lowke
//  see http://www.lowkemedia.com for more information
//
//  Permission is hereby granted, free of charge, to any person obtaining a 
//  copy of this software and associated documentation files (the "Software"), 
//  to deal in the Software without restriction, including without limitation 
//  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
//  and/or sell copies of the Software, and to permit persons to whom the 
//  Software is furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in 
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
//  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
//  IN THE SOFTWARE. 
//
//


package com.lowke.utils.as3.assetLoader
{   
    import com.lowke.utils.as3.Callback;
    import com.lowke.utils.as3.DoLater;
    import com.lowke.utils.as3.assetLoader.events.AssetLoadedEvent;
    import com.lowke.utils.as3.assetLoader.events.BatchCompleteEvent;
    import com.lowke.utils.as3.assetLoader.events.ProgressEvent;
    import com.lowke.utils.as3.assetLoader.loaders.AddSwf;
    import com.lowke.utils.as3.assetLoader.loaders.ILoad;
    import com.lowke.utils.as3.assetLoader.loaders.LoadBinary;
    import com.lowke.utils.as3.assetLoader.loaders.LoadClass;
    import com.lowke.utils.as3.assetLoader.loaders.LoadData;
    import com.lowke.utils.as3.assetLoader.loaders.LoadImage;
    import com.lowke.utils.as3.assetLoader.loaders.LoadSound;
    import com.lowke.utils.as3.assetLoader.loaders.LoadVariables;
    import com.lowke.utils.as3.assetLoader.loaders.LoadXML;
    import com.lowke.utils.as3.logAs.LogAs;
    
    import flash.display.Loader;
    import flash.events.Event;
    import flash.events.EventDispatcher;
    import flash.utils.Dictionary;
    import flash.utils.getDefinitionByName;
    
    // dispatched when all assets have completed loading
    [Event(name="assetLoader_batch_complete", type="com.lowke.utils.as3.assetLoader.events.BatchCompleteEvent")];
    
    // dispatched with a percentage progress report on all assets loading
    [Event(name="assetLoader_progress", type="com.lowke.utils.as3.assetLoader.events.ProgressEvent")];
    
    
    /**
     * AssetLoader Class</p>
     * 
     * AssetLoader is designed to allow you to easily bulk load assets from 
     * external files. In particular, it easily and conveniently supplies you with 
     * a single reference to your external asset. Clearing that reference generally 
     * will release the asset from memory (see note below). This tidies up the mess 
     * of event listeners and object references often associated with the Flash 
     * Loader that can make bulk loading of external assets and clearing of those 
     * assets tedious.
     *  
     * <p>AssetLoader is a singleton, a reference to it is obtained through
     * AssetLoader.instance, though static versions of all of its methods are 
     * supplied. See the AssetLoaderExample project for an example of how to use 
     * AssetLoader.</p>
     * 
     * <p>Note: Flash DisplayObject assets can be unusually stubborn to unload, 
     * particularly if the asset is a MovieClip that has children that are playing 
     * or if there are hard EventListeners attached to it. You should stop all 
     * MovieClips and remove all EventListeners on an asset before clearing the 
     * reference. As of Flash Player 10 you can dispose of stubborn DisplayObject 
     * assets using the Flash Player 10 unloadAndStop() method on the asset's Flash 
     * Loader, which can be accessed through loaderInfo.loader, like so,
     * <code>displayObjectAsset.loaderInfo.loader.unloadAndStop();</code></p>
     *  
     * <p>To get the progress of the AssetLoader while loading assets use,
     * 
     * <pre>
     *  import com.lowke.utils.as3.assetLoader.events.ProgressEvent;
     *  AssetLoader.addEventListener(AssetLoader.PROGRESS, printProgress);
     *
     *  public function printProgress(evt:ProgressEvent):void {
     *      trace("Percent loaded:" + Math.round(evt.percentDone*100) + "%");
     *  }
     * </pre>
     * 
     * You can also get the progress by querying the AssetLoader directly,
     * 
     * <pre>
     *  var percentDone:Number = AssetLoader.percentDone;
     * </pre>
     * 
     * <p/>Note: You might need to set permissions for local file access from the 
     * Flash Global Security Settings panel, see,
     * http://www.macromedia.com/support/documentation/en/flashplayer/help/
     * and select "Global Security Settings Panel"</p>
     * 
     * @author Russell J Lowke
     * @langversion ActionScript 3.0
     * @playerversion Flash 9
     * @see #loadFile()
     * @see http://www.lowkemedia.com
     */
    public class AssetLoader extends EventDispatcher
    {   
        //
        // error, warning and ifo IDs
        
        public static const WARNING_CANT_RECOGNIZE_FILE_TYPE:int    = 1000;
        public static const WARNING_CANT_RECOGNIZE_LOADER_TYPE:int  = 1001;
        public static const WARNING_FILE_NOT_FOUND:int              = 1002;
        public static const WARNING_FILE_CANT_LOAD:int              = 1003;
        public static const WARNING_CLASS_NOT_FOUND:int             = 1004;
        public static const WARNING_SECURITY_BREACH:int             = 1005;
        public static const WARNING_LOAD_TIMEOUT:int                = 1006;
        public static const INFO_LOADER_ISSUE:int                   = 1007;
        public static const INFO_FILE_CLOSED:int                    = 1008;
        public static const DEBUG_FILE_REQUESTED:int                = 1009;
        public static const DEBUG_FILE_LOADED:int                   = 1010;
        public static const DEBUG_BATCH_COMPLETE:int                = 1011;
        
        //
        // internal loader types
        public static const SWF:String       = "swf";           // .swf             Small Web Format/Flash
        public static const IMAGE:String     = "image";         // .bmp             Bitmap image file format
                                                                // .png             Portable Network Graphics
                                                                // .jpg, .jpeg      Joint Photographic Experts Group
        public static const SOUND:String     = "sound";         // .aif, .aiff      Audio Interchange File Format 
                                                                // .mp3             MPEG-1 Audio Layer 3
                                                                // .wav             Waveform audio format       
        public static const DATA:String      = "data";          // .html            Hypertext Markup Language
                                                                // .txt, .text      Text (ASCI) data
        public static const XML:String       = "xml";           // .xml             Extensible Markup Language
        public static const BINARY:String    = "binary";        // .bin             Binary
        public static const VARIABLES:String = "variables";     // .var, .vars      Url variables
        
        // public static const VIDEO:String  = "video";         //  (not currently supported)
                                                                // .flv             Flash Video,    
                                                                // .f4v             Video for Adobe Flash Player
                                                                // .f4p             Protected Video for Adobe Flash Player
                                                                // .f4a             Audio for Adobe Flash Player
                                                                // .f4b             Audio Book for Adobe Flash Player
        
        
        
        // static reference to AssetLoader singleton
        private static const INSTANCE:AssetLoader = new AssetLoader();
        
        // member variables
        private var _loaders:Dictionary = new Dictionary(true); // Keep a reference to loaders in use
        private var _nLoaders:uint = 0;                         // # of loaders currently in use
        private var _nFilesTotal:uint = 0;                      // # of files being loaded
        private var _percentDone:Number = NaN;                  // percent loaded of _nFilesTotal, NaN returned if idle

        /**
         * AssetLoader is a singleton and should never be constructed directly.
         * Use AssetLoader.instance to obtain a reference to the AssetLoader.
         */
        public function AssetLoader() 
        {
            super();
            
            if (INSTANCE != null) {
                 // show error to prevent new instances of AssetLoader being created.
                throw new Error("AssetLoader is a singleton, it should never be created twice.\r" + 
                                "Use AssetLoader.instance to get a reference to AssetLoader.");
            }
        }

        /**
         * loadFile()</p>
         * 
         * Loads an external file whose file type is determined by its file extension.
         * Once the asset has finished loading it is passed as the 1st argument to a function callback.
         * The type of the 1st parameter of the function callback should reflect the type of the file being loaded.</p>
         * 
         * For example, loading an XML file would be done like so,</p>
         * 
         * <pre>
         *  AssetLoader.loadFile("config.xml", gotXML);
         * </pre>
         * 
         * With the callback function looking like so,</p>
         * 
         * <pre>
         *  private function gotXML(cfg:XML):void {
         *      if (! cfg) {
         *          trace("XML failed to load");
         *      } else {
         *          trace("XML CONFIG:" + cfg);
         *      }
         *  }
         * </pre>
         * 
         * If an asset fails to load the callback function will still be called, only the asset being passed to it will 
         * be returned as null and a warning message will be logged looking something like so,
         * <pre>
         *  WARNING #1002: AssetLoader IOError, usually caused by file not found.
         *  Error #2035: URL Not Found. URL: file:///Users/russ/Workspace/AssetLoaderExample/bin-debug/config.xml
         * </pre>
         * 
         * <p>You can bulk load a series of assets by sending a series of consecutive loadFile() (or loadClass or similar) calls
         * concluding with a whenDone() call to a callback function.</p>
         * 
         * <pre>
         *  AssetLoader.loadFile("config.xml", gotAsset);
         *  AssetLoader.loadFile("myGraphic.jpg", gotAsset);
         *  AssetLoader.loadFile("mySound.mp3", gotAsset);
         *  AssetLoader.whenDone(ready);
         *  
         *  private function gotAsset(asset:Object):void {
         *      trace("Asset arrived:" + asset);
         *  }
         *  
         *  private function ready(evt:Event = null):void {
         *      trace("Your assets have loaded.");
         *  }
         * </pre>
         * 
         * 
         * Additional agruments can be sent to the loadFile callback to help identify the asset loaded.</p>
         * 
         * AssetLoader recognises the following file types:</p>
         * 
         * <table>
         * <tr> <th>SUFFIX</th>             <th>CLASS</th>              <th>FILE TYPE</th>  </tr>
         * <tr> <td>".swf"</td>             <td>Sprite/MovieClip</td>   <td>Shockwave Flash</td>    </tr>
         * <tr> <td>".bmp"</td>             <td>Bitmap</td>             <td>Bitmap image file format</td>   </tr>
         * <tr> <td>".png"</td>             <td>Bitmap</td>             <td>Portable Network Graphics</td>  </tr>
         * <tr> <td>".jpg", or ".jpeg"</td> <td>Bitmap</td>             <td>Joint Photographic Experts Group</td>   </tr>
         * <tr> <td>".aif", or ".aiff"</td> <td>Sound</td>              <td>Audio Interchange File Format</td>  </tr>
         * <tr> <td>".mp3"</td>             <td>Sound</td>              <td>MPEG-1 Audio Layer 3</td>   </tr>
         * <tr> <td>".wav"</td>             <td>Sound</td>              <td>Waveform Audio Format</td>  </tr>
         * <tr> <td>".xml"</td>             <td>XML</td>                <td>Extensible Markup Language</td> </tr>
         * <tr> <td>".plist"</td>           <td>XML</td>                <td>Mac OS X Property list</td> </tr>
         * <tr> <td>".bin"</td>             <td>ByteArray</td>          <td>Binary data</td>    </tr>
         * <tr> <td>".txt"</td>             <td>String</td>             <td>Text</td>   </tr>
         * <tr> <td>".html"</td>            <td>String</td>             <td>Hypertext Markup Language</td>  </tr>
         * <tr> <td>".var" or ".vars"</td>  <td>URLVariables</td>       <td>Url variables</td>  </tr>
         * </table>
         * 
         * 
         * Note: ".swf" files should be typecast as <code>Sprite</code>, if a single frame, or as 
         * <code>MovieClip</code>, if multiple frames.</p>
         * 
         * If you want to explicity declare the file type use the <code>load()</code> method instead.</p>
         * 
         * @param url URL path to asset being loaded. File type is determined by the file extension suffix in the url.
         * Typical file types are .jpg, .mp3, .swf, and .xml. 
         * @param funct Function callback which is passed the loaded asset once load has completed.
         * @param ...args Additional arguments sent to function callback after the loaded asset.
         * 
         * @return Returns an ILoad loader which can be used to stop the load early using close(), or listened to for 
         * events such as ProgressEvent.PROGRESS or AssetLoadedEvent.ASSET_LOADED. Remember to clear
         * any reference you keep to this loader otherwise any assets loaded by it will not clear from memory.
         * 
         * @see #load()
         */
        public function loadFile(url:String, 
                                 funct:Function = null, 
                                 ... args):ILoad 
        {   
            // determine loader type using file suffix
            var suffix:String = stripCGI(url);
            suffix = suffix.slice(suffix.lastIndexOf("."));
            var loaderType:String = loaderType(suffix);
            
            if (! loaderType) {
                LogAs.warning("AssetLoader can't recognize loader type for use loading file with suffix of " + suffix + " and url " + url + ".\r" + 
                              "Deafult image loader being used to load the file.", WARNING_CANT_RECOGNIZE_FILE_TYPE);
                loaderType = IMAGE;
            }
            
            return load.apply(null, [url, loaderType, funct].concat(args));
        }
        /** @see #loadFile() */
        public static function loadFile(url:String, funct:Function = null, ... args):ILoad 
        {
            return INSTANCE.loadFile.apply(null, [url, funct].concat(args));
        }
        
        
        /**
         * load()</p>
         * 
         * Loads external file as a particular file type using an explicit loader.</p>
         * Possible loader types are:
         * <table>
         * <tr> <th>LOADER TYPE</th>            <th>SUFFIX</th>         <th>FILE TYPE</th>  </tr>
         * <tr> <td>AssetLoader.SWF</td>        <td>.swf</td>           <td>Shockwave Flash</td>    </tr>
         * <tr> <td>AssetLoader.IMAGE</td>      <td>.bmp</td>           <td>Bitmap image file format</td>   </tr>
         * <tr> <td>AssetLoader.IMAGE</td>      <td>.png</td>           <td>Portable Network Graphics</td>  </tr>
         * <tr> <td>AssetLoader.IMAGE</td>      <td>.jpg, .jpeg</td>    <td>Joint Photographic Experts Group</td>   </tr>
         * <tr> <td>AssetLoader.SOUND</td>      <td>.aif, .aiff</td>    <td>Audio Interchange File Format</td>  </tr>
         * <tr> <td>AssetLoader.SOUND</td>      <td>.mp3</td>           <td>MPEG-1 Audio Layer 3</td>   </tr>
         * <tr> <td>AssetLoader.SOUND</td>      <td>.wav</td>           <td>Waveform Audio Format</td>  </tr>
         * <tr> <td>AssetLoader.XML</td>        <td>.xml</td>           <td>Extensible Markup Language</td> </tr>
         * <tr> <td>AssetLoader.XML</td>        <td>.plist</td>         <td>Mac OS X Property list</td> </tr>
         * <tr> <td>AssetLoader.DATA</td>       <td>.html</td>          <td>Hypertext Markup Language</td>  </tr>
         * <tr> <td>AssetLoader.DATA</td>       <td>.txt, .text</td>    <td>Text (ASCI) data</td>   </tr>
         * <tr> <td>AssetLoader.BINARY</td>     <td>.bin</td>           <td>Binary</td> </tr>
         * <tr> <td>AssetLoader.VARIABLES</td>  <td>.var, .vars</td>    <td>Url variables</td>  </tr>
         * </table>
         *  
         * 
         * 
         * @param url URL path to file being loaded.
         * @param loaderType loader type to use when loading the asset.
         * @param funct Function callback which is passed the loaded asset once load has completed.
         * @param ...args Additional arguments sent to function callback after the loaded asset.
         * 
         * @return Returns an ILoad loader which can be used to stop the load early using close(), or listened to for 
         * events such as ProgressEvent.PROGRESS or AssetLoadedEvent.ASSET_LOADED. Remember to clear
         * any reference you keep to this loader otherwise any assets loaded by it will not clear from memory.
         */
        public function load(url:String,
                             loaderType:String,
                             funct:Function = null,
                             ... args):ILoad
        {   
            // USE var assetLoader:ILoad = getLoader(url); TO CHECK FOR ASSET ALREADY LOADING?
            
            // use appropriate loader
            var loader:ILoad;
            switch (loaderType) {
                case SWF:       // SWF uses same loader as IMAGE        HAVE OWN LOADER?
                case IMAGE:     loader = new LoadImage(url);            break;
                case SOUND:     loader = new LoadSound(url);            break;
                case DATA:      loader = new LoadData(url);             break;
                case XML:       loader = new LoadXML(url);              break;
                case BINARY:    loader = new LoadBinary(url);           break;
                case VARIABLES: loader = new LoadVariables(url);        break;
                // case VIDEO:  loader = new LoadVideo(url);            break;
                
                default:
                
                    LogAs.warning("AssetLoader can't recognize loader type \"" + loaderType + "\" for url " + url + ".\r" +
                                  "Deafult image loader being used to load the file.", WARNING_CANT_RECOGNIZE_LOADER_TYPE);
                    loader = new LoadImage(url);
            }
            
            loadUsing.apply(null, [loader, funct].concat(args));
            return loader;
        }
        /** @see #load() */
        public static function load(url:String, loaderType:String = null, funct:Function = null, ... args):ILoad 
        {
            return INSTANCE.load.apply(null, [url, loaderType, funct].concat(args));
        }
        
        
        /**
         * loadUsing()</p>
         * 
         * Uses a specific ILoad object to load a file.
         * 
         * @param url URL path to file being loaded.
         * @param funct Function callback called when the asset has finished loading.
         * @param ...args Arguments sent to function callback.
         */
        public function loadUsing(loader:ILoad,
                                  funct:Function = null,
                                  ... args):void
        {   
            addAssetLoader(loader);
            Callback.callWithEvt(loader as EventDispatcher, AssetLoadedEvent.ASSET_LOADED, deliverAssetEvent, funct, args);
        }
        /** @see #loadUsing() */
        public static function loadUsing(loader:ILoad, funct:Function = null, ... args):void 
        {
            INSTANCE.loadUsing.apply(null, [loader, funct].concat(args));
        }
        
        
        /**
         * loadClass()</p>
         * 
         *  You may load a Class object out of a specific .swf file
         *  and send it to a function once it has loaded using loadClass().
         *
         *  This is achieved like so,
         * <pre>
         *      AssetLoader.loadClass("assest.swf", "ClassName", gotAsset);
         *
         *      private function gotAsset(asset:Object):void {
         *          trace("Asset arrived:" + asset);
         *      }
         * </pre>
         * 
         * With loadClass() a new instance of the class object is given to the callback function,
         * if you want the class definition you should use loadClassDefinition().
         *
         *  Note: calls to multiple classes in the same .swf in the same batch will not cause the
         *  .swf to be loaded multiple times, only once.
         * 
         * @param url URL path to .swf file being loaded.
         * @param className Name of class being loaded from .swf file.
         * @param funct Function callback called when the asset has finished loading.
         * @param ...args Arguments sent to function callback.
         * 
         * @return Returns an ILoad loader which can be used to stop the load early using close(), or listened to for 
         * events such as ProgressEvent.PROGRESS or AssetLoadedEvent.ASSET_LOADED. Remember to clear
         * any reference you keep to this loader otherwise any assets loaded by it will not clear from memory.
         */
        public function loadClass(url:String,
                                  className:String,
                                  funct:Function = null,
                                  ... args):ILoad 
        {                           
            return classLoad(url, className, false, funct, args);
        }
        /** @see #loadClass() */
        public static function loadClass(url:String, className:String, funct:Function = null, ... args):ILoad 
        {
            return INSTANCE.loadClass.apply(null, [url, className, funct].concat(args));
        }
        
        
        /**
         * loadClassDefinition()</p>
         * 
         * Works the same as loadClass() except returns the class definition from the .swf
         * rather than an instance of the class.
         *  Note: calls to multiple classes in the same .swf in the same batch will not cause the
         *  .swf to be loaded multiple times, only once.
         * 
         * @param url URL path to .swf file being loaded.
         * @param className Name of class being loaded from .swf file.
         * @param funct Function callback called when the asset has finished loading.
         * @param ...args Arguments sent to function callback.
         * 
         * @return Returns an ILoad loader which can be used to stop the load early using close(), or listened to for 
         * events such as ProgressEvent.PROGRESS or AssetLoadedEvent.ASSET_LOADED. Remember to clear
         * any reference you keep to this loader otherwise any assets loaded by it will not clear from memory.
         *
         * @see #loadClass()
         */
        public function loadClassDefinition(url:String,
                                            className:String,
                                            funct:Function,
                                            ... args):ILoad 
        {
            return classLoad(url, className, true, funct, args);
        }
        /** @see #loadClassDefinition() */
        public static function loadClassDefinition(url:String, className:String, funct:Function, ... args):ILoad 
        {
            return INSTANCE.loadClassDefinition.apply(null, [url, className, funct].concat(args));
        }
        
        
        /**
         * addSwf()</p>
         * 
         * Adds a .swf file to the application domain so all
         * its internal Classes may be accessed by the application using 
         * getClassInstance(). 
         * 
         * <p>Careful using addSwf() as it's difficult to clear
         * classes once they are in the application domain.
         * Usually you're better off using loadClass()</p>
         * 
         * @param url URL path to .swf file being added to the application domain.
         * @param funct Function callback called when the asset has finished loading.
         * @param args Arguments sent to function callback.
         * 
         * @return Returns an ILoad loader which can be used to stop the load early using close(), or listened to for 
         * events such as ProgressEvent.PROGRESS or AssetLoadedEvent.ASSET_LOADED. Remember to clear
         * any reference you keep to this loader otherwise any assets loaded by it will not clear from memory.
         */
        public function addSwf(url:String,
                               funct:Function = null, 
                               ... args):ILoad
        {   
            var assetLoader:ILoad = new AddSwf(url);
            Callback.callWithEvt(assetLoader as EventDispatcher, AssetLoadedEvent.ASSET_LOADED, deliverAssetEvent, funct, args);
            addAssetLoader(assetLoader);
            return assetLoader;
        }
        /** @see #addSwf() */
        public static function addSwf(url:String, funct:Function = null, ... args):ILoad {
            return INSTANCE.addSwf.apply(null, [url, funct].concat(args));      
        }
        
        
        /**
         * getClassInstance()</p>
         * 
         * A shortcut method that uses getDefinitionByName() 
         * to retrieve a new instance of a Class from the application domain.
         * 
         * @param className Class name in the application domain.
         * 
         * @return Instance of class specified by className.
         */
        public static function getClassInstance(className:String):* 
        {
            var mClass:Class = getDefinitionByName(className) as Class;
            return new mClass();
        }
        
        
        /**
         * close()</p>
         * 
         * Closes a file being loaded.</p>
         * 
         * To stop an asset from loading call close(), passing it the url path
         * to the asset being loaded.
         * 
         * @param url URL path to file to close.
         */
        public function close(url:String):void
        {   
            var loader:ILoad = getLoader(url);
            if (loader) {
                loader.close();
            }
        }
        /** @see #close() */
        public static function close(url:String):void {
            INSTANCE.close(url);
        }

        
        /**
         * getLoader()</p>
         * 
         * Gets a reference to an ILoader loader currently being used to load a file.
         * 
         * @param url URL path to file being loaded
         * @param loaderClass can be used to specify a specific ILoad class type of the loader being requested.
         * 
         * @return Returns a reference to an in-progress ILoad loader for a file which can be used to stop
         * the load early using close(), or listened to for events such as ProgressEvent.PROGRESS or 
         * AssetLoadedEvent.ASSET_LOADED. Remember to clear any reference you keep to this loader otherwise 
         * any assets loaded by it will not clear from memory.
         */
        public function getLoader(url:String, 
                                  loaderClass:Class = null):ILoad 
        {   
            // check if url file already in current loading cue...
            url = stripCGI(url);
            for each(var loader:ILoad in _loaders) {
                if (url == stripCGI(loader.url)) {
                    
                    // check if it is the right type of loader
                    if (loaderClass) {
                        if (loader is loaderClass) {
                            return loader;
                        }
                    } else {
                        return loader;
                    }
                }
            }
            return null;
        }
        /** @see #getLoader() */
        public static function getLoader(url:String, loaderClass:Class = null):ILoad 
        {
            return INSTANCE.getLoader(url, loaderClass);
        }

        
        /**
         * whenDone()</p>
         * 
         * A convenience method that sets a function callback to be called 
         * when all current loads have finished loading. Multiple consecutive whenDone() 
         * callbacks may be set. 
         * 
         * <p>If the AssetLoader is idle with no loads running when whenDone() is 
         * called then the whenDone() function callback will be called on the next frame,
         * when Flash sends a general Event.ENTER_FRAME.</p>
         * 
         * <p>whenDone() listens for the BatchCompleteEvent.BATCH_COMPLETE event to be sent 
         * from AssetLoader.instance</p>
         * 
         * @param funct Function callback called when batch finished loading.
         * @param ...args Arguments sent to function callback.
         */
        public function whenDone(funct:Function, ... args):void
        {   
            if (idle) {
                // use DoLater to trigger callback on the next frame
                new DoLater(funct, args);
            } else {
                Callback.argsCall(this, BatchCompleteEvent.BATCH_COMPLETE, funct, args);
            }
        }
        /** @see #whenDone() */
        public static function whenDone(funct:Function, ... args):void 
        {
            INSTANCE.whenDone.apply(null, [funct].concat(args));
        }
        
        //
        // Methods called by internal loaders, etc
        //
        
        //
        // add a loader to _loaders
        private function addAssetLoader(assetLoader:ILoad):void
        {   
            if (_nFilesTotal == 0) {
                // dispatch immediate initial progress event
                dispatchEvent(new ProgressEvent(ProgressEvent.PROGRESS, 0));
            }
            ++_nFilesTotal;
            
            // keep _loader reference safe from garbage collection
            _loaders[assetLoader] = assetLoader;
            ++_nLoaders;
        }
        
        
        /**
         * @private this method is for internal use only. 
         * If you need to remove a loader use close().
         * 
         * Removes exiting AssetLoader from _loaders queue.
         */
        public function removeAssetLoader(assetLoader:ILoad):void
        {   
            assetLoader.dispose();
            
            // and remove loader from _loaders
            _loaders[assetLoader] = null;
            delete _loaders[assetLoader];
            --_nLoaders;
            
            // if all loaders finished
            if (_nLoaders == 0) {
                
                // reset counters
                _nFilesTotal  = 0;
                _percentDone  = NaN;
                
                LogAs.debug("AssetLoader batch complete.", AssetLoader.DEBUG_BATCH_COMPLETE);
                dispatchEvent(new BatchCompleteEvent(BatchCompleteEvent.BATCH_COMPLETE));
            }
        }
        
        
        //
        // private method to deal with
        //  loadClass() and loadClassDefinition()
        private function classLoad(url:String,
                                   className:String,
                                   giveClassDef:Boolean,
                                   funct:Function,
                                   args:Array):ILoad
        {   
            // check if swf file path already in current loading queue...
            var assetLoader:ILoad = getLoader(url, LoadClass);
            
            if (! assetLoader) {
                // otherwise create a new class Loader
                assetLoader = new LoadClass(url);
                addAssetLoader(assetLoader);
            }
            
            // listen for asset to load
            Callback.callWithEvt(assetLoader as EventDispatcher, 
                AssetLoadedEvent.ASSET_LOADED, 
                deliverClassEvent, 
                className, 
                giveClassDef, 
                funct, args);
            return assetLoader;
        }
        
        //
        // private method used to deliver a class by digging
        //  the Class definition out of the .swf
        private function deliverClassEvent(evt:AssetLoadedEvent,
                                           className:String,
                                           giveClassDef:Boolean,
                                           funct:Function,
                                           args:Array):void
        {
            // retrieve Loader from event
            var loader:Loader = evt.asset;
            
            // dig the Class definition out of the .swf
            var classDef:Class;
            var asset:*;

            if (loader) {
                
                try {
                    classDef = loader.contentLoaderInfo.applicationDomain.getDefinition(className) as Class;
                    
                    // assign asset from class definition
                    asset = giveClassDef ? classDef : new classDef();
                    
                } catch (err:Error) {
                    LogAs.warning("AssetLoader could not find Class \"" + className + "\" in file " + evt.url + "\r" + err.toString(), AssetLoader.WARNING_CLASS_NOT_FOUND);
                }
            }
            
            deliverAsset(asset, funct, args);
        }
        
        //
        // private methods used to deliver an asset to the function callback
        private function deliverAssetEvent(evt:AssetLoadedEvent,
                                         funct:Function, 
                                         args:Array):void 
        {
            deliverAsset(evt.asset, funct, args);
        }
        private function deliverAsset(asset:*, 
                                      funct:Function, 
                                      args:Array):void
        {   
            // if function to deliver to
            if (funct != null) {

                // trigger function, adding asset to beginning of argument list
                funct.apply(null, [asset].concat(args));
            }
        }
        
        
        //
        // dispatches update progress of AssetLoader
        public function dispatchProgress(evt:Event = null):void
        {
            var progress:Number = progress();
            if (_percentDone != progress) {
                _percentDone = progress;
                dispatchEvent(new ProgressEvent(ProgressEvent.PROGRESS, _percentDone));
            }
        }
        
        //
        // iterates through active loaders calculating total progress
        private function progress():Number
        {   
            if (_nFilesTotal == 0) {
                return NaN;
            }
            
            // iterate through each loader getting percent done
            var sofar:Number = _nFilesTotal - _nLoaders;
            for each(var loader:ILoad in _loaders) {
                var loaderPercentDone:Number = loader.percentDone;
                if (isNaN(loaderPercentDone)) {
                    loaderPercentDone = 0;
                }
                sofar += loaderPercentDone;
            }

            var percentDone:Number = sofar/_nFilesTotal;
            if (percentDone > 1) {
                percentDone = 1;
            }
            return percentDone; 
        }
        
        
        //
        // static helper methods
        //

        
        /**
         * Strips any CGI "?" variables from a url String
         *
         * @url URL String being stripped
         * 
         * @return Returns the cleaned URL String
         */
        public static function stripCGI(url:String):String
        {
            var index:int = url.indexOf('?');
            if (index > -1) {
                // strip out any "?" CGI data
                url = url.slice(0, index);
            }
            return url;
        }
        

        /**
         * @param suffix File suffix such as <code>".jpg"</code> indicating file type.
         * 
         * @return Returns AssetLoader constant indicating file type. <code>null</code> returned if suffix not recognized.
         */
        public static function loaderType(suffix:String):String 
        {   
            suffix = suffix.toLowerCase();

            // find appropriate loader from suffix
            switch (suffix) {
                
                case ".swf":                        // Shockwave Flash/Small Web Format
                    return SWF;
                
                case ".aif": case ".aiff":          // Audio Interchange File Format 
                case ".mp3":                        // MPEG-1 Audio Layer 3
                case ".wav":                        // Waveform audio format
                    return SOUND;
                
                case ".bmp":                        // Bitmap image file format
                case ".png":                        // Portable Network Graphics
                case ".jpg": case ".jpeg":          // Joint Photographic Experts Group
                    return IMAGE;
                
                case ".html":                       // Hypertext Markup Language
                case ".txt": case ".text":          // Text (ASCI) data
                    return DATA;
                    
                case ".xml":                        // Extensible Markup Language
                case ".plist":                      // plist xml data
                    return XML;
                
                case ".bin":                        // binary
                    return BINARY;
                
                case ".var": case ".vars":          // url variables
                    return VARIABLES;
                
                //
                // (Video not yet supported)
                //
                // case ".flv":                 // Flash Video, (not currently supported)
                // case ".f4v":                 // Video for Adobe Flash Player
                // case ".f4p":                 // Protected Video for Adobe Flash Player
                // case ".f4a":                 // Audio for Adobe Flash Player
                // case ".f4b":                 // Audio Book for Adobe Flash Player
                //  return VIDEO;
                
                default:
                    return null;            
            }
        }
        
        
        //
        // accessors and mutators
        //
        
        /** Gives the percentage loaded of the current batch of loading files, where 1% is returned as 0.01 */
        public function get percentDone():Number                { return _percentDone; }
        
        /** Indicates if the AssetLoader is idle with no loads being performed. */
        public function get idle():Boolean                      { return (_nFilesTotal == 0); }
        
        /** Gives a reference to the AssetLoader singleton. */
        public static function get instance():AssetLoader       { return INSTANCE; }
        
        /** @see #percentDone */
        public static function get percentDone():Number         { return INSTANCE.percentDone; }
        
        /** @see #idle */
        public static function get idle():Boolean               { return INSTANCE.idle; }
    }
}
