///
//  LoadBase v 1.3
//  Russell Lowke, January 18th 2011
//
//  Copyright (c) 2008-2011 Russell Lowke
//  see http://www.lowkemedia.com for more information
// 
//  Permission is hereby granted, free of charge, to any person obtaining a 
//  copy of this software and associated documentation files (the "Software"), 
//  to deal in the Software without restriction, including without limitation 
//  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
//  and/or sell copies of the Software, and to permit persons to whom the 
//  Software is furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in 
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
//  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
//  IN THE SOFTWARE. 
//
//

package com.lowke.utils.as3.assetLoader.loaders 
{   
    import com.lowke.utils.as3.Delay;
    import com.lowke.utils.as3.assetLoader.AssetLoader;
    import com.lowke.utils.as3.assetLoader.events.AssetLoadedEvent;
    import com.lowke.utils.as3.assetLoader.events.LoaderDisposedEvent;
    import com.lowke.utils.as3.logAs.LogAs;
    
    import flash.events.Event;
    import flash.events.EventDispatcher;
    import flash.events.IOErrorEvent;
    import flash.events.SecurityErrorEvent;
    import flash.net.URLRequest;
    
    // dispatched when an asset has loaded
    [Event(name="assetLoader_assetLoaded", type="com.lowke.utils.as3.assetLoader.events.AssetLoadedEvent")];
    
    // dispatched with loader disposed of, either due to being closed, failing, or finishing loading
    [Event(name="assetLoader_loader_disposed", type="com.lowke.utils.as3.assetLoader.events.LoaderDisposedEvent")];
    
    
    public class LoadBase extends EventDispatcher implements ILoad 
    {
        private static var idCount:uint = 0;        // used to create unique ids for AssetLoaders
        
        //
        // member variables
        protected var _assetLoader:AssetLoader;     // refrenece to asset loader
        protected var _request:URLRequest;          // URLRequest to the asset being loaded
        protected var _loader:Object;               // Flash loader either, Loader(), URLLoader(), or Sound()
        private var _timeoutRetry:uint = 10000;     // after 10 seconds loader will timeout and try again.
        private var _nAttempts:uint;                // number of attempts made to load the item
        protected var _id:uint;                     // each loader is given a unique numeric id.
        private var _delay:Delay;                   // timer used to ensure loader has not timed out
        
        //
        // constructor
        public function LoadBase(url:String) 
        {
            _assetLoader = AssetLoader.instance;
            _request = new URLRequest(url);
            _nAttempts = 1;
            _id = idCount++;
            _delay = Delay.doIt(_timeoutRetry, timeoutError);
            LogAs.debug("File requested (" + _id + "): " + _request.url, AssetLoader.DEBUG_FILE_REQUESTED);
            createLoader();
        }
        
        //
        // close the loader before it gets a chance to finish loading
        public function close():void 
        {
            LogAs.info("AssetLoader closing loading file (" + _id + "):" + this.url, AssetLoader.INFO_FILE_CLOSED);
            
            // Note: removeAssetLoader() will call dispose() on this instance
            _assetLoader.removeAssetLoader(this);
        }
        
        public function dispose(retry:Boolean = false):void 
        {
            // dispose of this loader, called by removeAssetLoader() in AssetLoader
            _delay.close();
            try {
                // Note: This close() will try to close streaming sounds and streaming video,
                //  in cases where no stream is found an error can be thrown.
                _loader.close();
            } catch(err:Error) {};
            cleanup();
            
            if (! retry) {
                dispatchEvent(new LoaderDisposedEvent(LoaderDisposedEvent.LOADER_DISPOSED));
            }
        }
        
        private function reload():void 
        {
             ++_nAttempts;
            dispose(true);
            _delay = Delay.doIt(_timeoutRetry, timeoutError);
            createLoader();
        }
        
        //
        // all failed loads are classed as WARNINGS
        protected function loadFailed(errStr:String, errID:int):void 
        {   
            // give warning that load failed
            LogAs.warning(errStr, errID);
            
            // dispatch asset as null
            dispatchEvent(new AssetLoadedEvent(AssetLoadedEvent.ASSET_LOADED, null, _request.url));
            
            // remove loader
            _assetLoader.removeAssetLoader(this);
        }
        
        
        protected function ioErrorHandler(evt:IOErrorEvent):void 
        {    
            var errStr:String = evt.text;
            
            // can't access evt.errorID,
            //  so will use split() on error text instead.
            var errID:String = errStr.split(':', 2)[0];
            
            switch (errID) {
                
                case "Error #2036":                 // Error #2036: Load Never Completed.
                case "Error #2032":                 // Error #2032: Stream Error. Flash web services gives Stream Error instead of URL Not Found.
                case "Error #2035":                 // Error #2035: URL Not Found.  
                    loadFailed("AssetLoader IOError (attempt " + _nAttempts + "), usually caused by file not found.\rERROR: " + errStr, AssetLoader.WARNING_FILE_NOT_FOUND);
                    return;
                    
                default:
                    // otherwise show issue, but continue operation
                    LogAs.info("AssetLoader IOError (attempt " + _nAttempts + ")\rERROR: " + errStr, AssetLoader.INFO_LOADER_ISSUE);
                    break;
            }
            
            if (_nAttempts == 3) {
                loadFailed("AssetLoader IOError (attempt " + _nAttempts + ")\rERROR:" + errStr, AssetLoader.WARNING_FILE_CANT_LOAD);
                return;
                
            } else {
                
                // try to reload item
                reload();
            }
        }
        
        protected function securityErrorHandler(evt:SecurityErrorEvent):void 
        {
            // security issue
            loadFailed("AssetLoader Security Error:" + evt.text, AssetLoader.WARNING_SECURITY_BREACH);
        }
        
       
        private function timeoutError(prevPercentDone:Number = 0):void 
        {   
            // loader taking too long to load, might not be responding.
            
            var percentDone:Number = this.percentDone;
            if (percentDone > prevPercentDone) {
                
                // file is still loading, check again later
                _delay = Delay.doIt(_timeoutRetry, timeoutError, percentDone);
            } else {
                
                // otherwise loader not responding, retry
                LogAs.warning("AssetLoader Timeout Error (attempt " + _nAttempts + "). No response after " + _timeoutRetry + " milliseconds when loading " + _request.url, AssetLoader.WARNING_LOAD_TIMEOUT);
                reload();
            }
        }
        
        // createLoader is overwritten by other AssetLoader types
        public function createLoader():void 
        {
            throw new Error("createLoader() must be overridden");
        }
        
        // createLoader is overwritten by other AssetLoader types
        public function fileLoaded(evt:Event):void 
        {
            throw new Error("fileLoaded() must be overridden");
        }
        
        // prepare this object to be removed from memory
        public function cleanup():void 
        {
            throw new Error("cleanup() must be overridden");
        }
        
        //
        // return the percent done
        //  if nothing loaded percentDone may return NaN instead of 0, depending on specifics of the loader.
        public function get percentDone():Number 
        {
            throw new Error("percentDone() must be overridden");
        }
        
        
        //
        // accessors and mutators
        //
        
        public function get id():uint                       { return _id; }
        public function get url():String                    { return _request.url; }
        public function get timeoutRetry():uint             { return _timeoutRetry; }
        
        public function set timeoutRetry(val:uint):void     { _timeoutRetry = val; }
        
    }
}   